# -*- coding: utf-8 -*-
"""
 
 This file is part of oe3d (www.oe3d.at).

 oe3d is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 oe3d is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with oe3d.  If not, see <http://www.gnu.org/licenses/>.

 Author(s): Gerhard Summer (gerhard.summer@rechenraum.com) 
            Simon Flöry (simon.floery@rechenraum.com)
            

"""

# The code in here is based on
#	http://docs.cython.org/src/tutorial/numpy.html
# For a basic cython tutorial, see
#	http://docs.cython.org/src/tutorial/cython_tutorial.html

# Cython implementation of the core convolution in numberechnung. 
# See http://docs.cython.org/. The speed up is from 8 minutes down to 15 seconds
# per 1x1° ASTER tile
#
# 1. See compile_cython.py for compilation instructions. Compilation generates
#		numberechnung.so, that may be imported as module
# 2. Add 
#		import numberechnung"
# 	and call 
#		numberechnung.fenster(num, numneu, fenstergr)
# 	to use all this.

import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.float
ctypedef np.float_t DTYPE_t

# This "decoration" must be located immedeately before a function declaration
@cython.boundscheck(False)
@cython.cdivision(True)
def fenster(np.ndarray[DTYPE_t, ndim=2] num, np.ndarray[DTYPE_t, ndim=2] numneu, int fenstergr) :
	# All cdefs must be at the top intendation level
	cdef int i, j, k, l
	cdef DTYPE_t a, relnum
	
	# iterate over all entries of num and numneu. range(.) is automatically
	# converted to a proper C loop
	for i in range(3601):
		for j in range(3601):
			# compute maximum in a rectangular window around num[i,j]
			a = 0
			for k in range(max((i-fenstergr),0),min(3601,(i+fenstergr+1))):
				for l in range(max((j-fenstergr),0),min(3601,(j+fenstergr+1))):
					# cast to unsigned int to avoid checking for negative indices
					a=max(a,num[<unsigned int> k,<unsigned int> l])
			if a == 0:
				numneu[i,j]=0
			else:
				relnum=num[i,j]/a
				if relnum>1/3:
					# a parabola with f(1/3) = 0, f(1/2) = 0.1, f(1) = 1.
					# If you change this, make sure, that your parabola does not
					# evaluate to negative values!
					numneu[i,j]=1.8*relnum*relnum-0.9*relnum+0.1
				else:
					numneu[i,j]=0
	     
    #if num<5 numneu shall be zero, as we don't trust such small num-values. The numneu-values
    #of original values 5,6 and 7 shall be reduced, in order that the difference from 4 to 5 resp. 6
    #is not so sharp.      
	numneu[num==7]*=0.75
	numneu[num==6]*=0.5
	numneu[num==5]*=0.25
	numneu[num<5]=0
