# -*- coding: utf-8 -*-
"""

 This file is part of oe3d (www.oe3d.at).

 oe3d is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 oe3d is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with oe3d.  If not, see <http://www.gnu.org/licenses/>.

 Author(s): Gerhard Summer (gerhard.summer@rechenraum.com) 
            Simon Flöry (simon.floery@rechenraum.com)
            

"""

import warnings
import scipy
import numpy as np
import scipy.sparse.linalg as spla
from scipy.sparse import *
from scipy import *
from scipy.spatial import *
import time
import re
import os
import gdal
from gdalconst import *
import struct 
import osr
import datetime
import matplotlib.pyplot as plt
import math
import ConfigParser
import ctypes


################################################################################
#
# General initialization (in part from configuration file)
#
################################################################################


Config = ConfigParser.ConfigParser()
Config.read('./configfile.txt')

asterpfad = Config.get("dem", "asterpfad")
asternumpfad = Config.get("num", "numausgabepfad")
srtmpfad = Config.get("dem", "srtmpfad")
bevpfad = Config.get("dem", "bevpfad")
resultatpfad = Config.get("dem", "resultatpfad")
waterpfad = Config.get("dem", "waterpfad")

if (Config.has_option("dem", "objexportpfad")):
    objexportpfad = Config.get("dem", "objexportpfad")
else:
    objexportpfad = ''

if (Config.has_option("dem", "solver")):
    solver = Config.get("dem", "solver")
else:
    solver = 'direct_superlu'
    
if (Config.has_option("dem", "maxNrIterations")):
    maxNrIterations = Config.getint("dem", "maxNrIterations")
else:
    maxNrIterations = 5

# BEVWeight controls the influence of the BEV reference points:
#    = 1 ... The BEV points have the same influence as ASTER/SRTM
#    < 1 ... The BEV points have less influence than ASTER/SRTM
#    > 1 ... The BEV points have more influence than ASTER/SRTM
#    = 0 ... the BEV points are ignored for hoehenausgleich
if (Config.has_option("dem", "BEVweight")):
    BEVweight = Config.getfloat("dem", "BEVweight")
else:
    BEVweight = 1.

#if (Config.has_option("dem", "ASTERweight")):
#    ASTERweight = Config.getfloat("dem", "ASTERweight")
#else:
#    ASTERweight = 1.
# factor to get approximately the same ratio between longitudal and latitudal
# direction in WGS84 coordinate space as in real-world.
isofaktor=cos(47.5*2*pi/360)

# To work-around numerical issues, don't test floating point numbers for 
# equality but test the absolute difference vs. eps.
eps = 1e-8

# Label of water pixels in water mask
waterLabel = 16

# One place to set the sparse matrix storage type. Possible values: 'csc', 'csr'
sparse_matrix_type = 'csr'
#sparse_matrix_type = 'csc'


################################################################################


def compute(X0,Y0,dx,dy,lam,name):
    """
    This method finds all ASTER-,SRTM- and BEV-Files that cover a certain area
    (specified by X0,Y0, coordinates of the lower left corner, and dx,dy,i.e. length
    and width of the area). The data of these files are stored in a n x 3-array. 
    Finally a least squares adjustment is computed.
    
    dx and dy are a measure of length (in our own special unit). They are NOT
    the number of pixels. However, nrPixelsX = dx*10000+1 and 
    nrPixelsY = dy*10000+1
    """    
    startZeit = time.time()
    
    print('')
    print('================================================================================')
    print('oe3d (' + time.strftime("%d.%m.%Y %H:%M:%S") + ')')
    print('')
    print('parameters: (' + str(X0) + ', ' + str(Y0) + ') with dx=' + str(dx) + ' and dy=' + str(dy) + ', lambda = ' + str(lam) + '.')
    print('================================================================================')
    print('')
    
    #
    # read SRTM data
    #
    
    SRTMfiles=filefinder_ASTER_SRTM(X0,Y0,dx,dy)
    print('Importing SRTM data:')
    liste=[]
    for i in range(len(SRTMfiles)):
        print(' * Loading ' + str(SRTMfiles[i]) + ' ...')
        data=fileloader(srtmpfad+SRTMfiles[i]+'.hgt')
        data = data[data[:,2] != -32768,:]     
        liste.append(haircut(X0,Y0,dx,dy,data))
    if len(liste)>0:
        SRTM=np.vstack(liste)
        SRTM=unique(SRTM)
    else:
        SRTM=[]
    
    
    #
    # read ASTER data
    #
    
    ASTERfiles=filefinder_ASTER_SRTM(X0,Y0,dx,dy)
    print('Importing ASTER data')
    liste=[]
    for i in range(len(ASTERfiles)):
        print(' * Loading ' + str(ASTERfiles[i]) + ' ...')
        data=fileloader(asterpfad+ASTERfiles[i]+'.hgt')
        if asternumpfad != '':
             filename = ASTERfiles[i]+'_numneu.tif'
             print(' * Loading num-Datei ' + filename)
             num=fileloader(asternumpfad + filename)
             num=num[:,2]        
        else:
             num=32767*np.ones((len(data),1))
        numneu=num/32767.        
        data = np.c_[data,numneu]
        if os.path.exists(waterpfad+ASTERfiles[i]+'.tif'):
            filename = ASTERfiles[i]+'.tif';
            print(' * Loading Wassermaske ' + filename)
            gewaesser=fileloader(waterpfad + filename)   
            gewaesser=gewaesser[:,2]
        else:
            gewaesser=np.zeros((len(data),1))
        data = np.c_[data,gewaesser]
        liste.append(haircut(X0,Y0,dx,dy,data))
        
        
    if len(liste)>0:
        ASTER=np.vstack(liste)
        ASTER=unique(ASTER)
        indices=multi_indexing(X0,Y0,dx,dy,ASTER)
        ASTER = ASTER[indices,:]        
        
    else:
        ASTER=[]

    #
    # read BEV data
    #
    

    BEVfiles=filefinder_BEV(X0,Y0,dx,dy)
    print('Importing BEV data')
    liste=[]
    for i in range(len(BEVfiles)):
        print(' * Loading ' + str(BEVfiles[i]) + ' ...')
        data=fileloader(bevpfad+BEVfiles[i])    
        data = data[data[:,2] != -9999,:] 
        liste.append(haircut(X0,Y0,dx,dy,data))
    if len(liste)>0:
        BEV=np.vstack(liste)
        BEV=unique(BEV)
    else:
        BEV=[]
        
        
    print('Working with ' + str(SRTM.shape[0]) + ' SRTM, ' + str(ASTER.shape[0]) + ' ASTER and ' + str(shape(BEV)[0]) + ' BEV points.')
    print('That is a total of ' + str(SRTM.shape[0] + ASTER.shape[0] + shape(BEV)[0]) + ' points.')

    #
    # compute objective for closeness to elevation data
    #
    
    print('Setting up matrix/vector for closeness to given elevation data ...');
    time0 = time.time()
    (Ah,bh)=hoehenausgleich(X0,Y0,dx,dy,SRTM,ASTER,BEV)
    
    print('    Done in ' + str(time.time() - time0) + ' seconds.');
    
    #
    # compute objective for a smooth solution
    #
    
    print('Setting up matrix/vector for a smooth solution ...');
    time0 = time.time()
    (Ag,bg)=glaettung(X0,Y0,dx,dy,ASTER,BEV)
    
    print('    Done in ' + str(time.time() - time0) + ' seconds.');
    vergangenezeit = time.time()-startZeit
    print('Vergangene Zeit (Laden + Aufstellen Matrizen): ' + str(vergangenezeit))
    
    #
    # combine objectives and solve the least-squares problem
    #

    ST=time.time()
    
    A = scipy.sparse.vstack([Ah, lam*Ag], format=sparse_matrix_type)
    b = np.hstack((bh, lam*bg))
    
    Wi= scipy.sparse.eye(shape(A)[0], format=sparse_matrix_type)
    
    # constants for IRLS 
    irls_p=1.5
    irls_c=1.3998   

    z = np.zeros(A.shape[1]);
        
    for i in range(maxNrIterations):
        print('Iteration ' + str(i+1) + ' (of ' + str(maxNrIterations) + ')')

        Ai = Wi * A
        bi = Wi * b        
        
        if (solver == 'direct_superlu') : 
            z = solve_direct(Ai, bi, False)
        elif (solver == 'direct_umfpack') : 
            z = solve_direct(Ai, bi, True)
        elif (solver == 'iterative_cg') : 
            z = solve_cg(Ai, bi)
        elif (solver == 'iterative_qr') : 
            z = solve_qr(Ai, bi)
        elif (solver == 'iterative_mr') : 
            z = solve_mr(Ai, bi)
        else:
            print('ERROR: Unknown solver option value: ' + solver)

        ri = A * z - b 
             
        #diagonal = np.power(np.abs(ri), irls_p - 2)
        diagonal = 1/(1+np.abs(ri)/irls_c)
                
        # let IRLS work on the hoehenausgleich only
        diagonal[Ah.shape[0]:] = 1
        
        Wi = scipy.sparse.diags(diagonal, 0, format=sparse_matrix_type)
           
        residuumsum = np.dot(Wi * A * z - Wi * b, Wi * A * z - Wi * b)
        print('    res = ' + str(residuumsum))
                  
        
    print('Solving least-squares problem with ' + solver + ': ' + str(time.time()-ST))    
        
    #
    # we are done. Write result to disk.
    #
    
    fileexporter(X0,Y0,dx,dy,z[0:len(ASTER)],resultatpfad + name)
    
    # We are not interested in the values in the BEV points.
    if (objexportpfad != ''):
        fileexporter_obj(dx,dy,ASTER,z,objexportpfad+name+'.obj')  
        
    print('Vergangene Zeit (gesamt): ' + str(time.time() - startZeit))
    
    return(z)
    

################################################################################


def solve_direct(A, b, use_umfpack) :
    """
    Solve linear least-squares problem's normal equations with linear solver.
    If you run out of memory with the direct methods, switch to an iterative 
    method.
    
    A         One linear least-squares term per row, of size (Mh + Mg) x N, 
            with N the number of unknown elevation values and Mh the number of
            elevation terms and Mg the number of smoothing terms
    b         Right hand side, of size (Mh + Mg)
    use_umfpack        If true, UMFPACK is use to solve the linear system. 
                    Otherwise, SuperLU is used.
    """

    # build normal equations
    b = A.T * b
    A = A.T * A
    
    # solve and return
    return spla.spsolve(A, b, use_umfpack=use_umfpack)
    

################################################################################


def solve_cg(A, b) :
    """
    Solve linear least-squares problem's normal equations by CG iteration
    
    A         One linear least-squares term per row, of size (Mh + Mg) x N, 
            with N the number of unknown elevation values and Mh the number of
            elevation terms and Mg the number of smoothing terms
    b         Right hand side, of size (Mh + Mg)
    """
    
    # See https://github.com/pyamg/pyamg-examples/tree/master/SWIG for some
    # other preconditioners 
    #
    # see also https://github.com/dpo/pykrylov/tree/master/pykrylov
    #
    # scilab has preconditioners, and python bindings of scilab do exist
    
    # build normal equations
    b = A.T * b
    A = A.T * A
    
    # It would be better to use some incomplete Cholesky factorization as A is
    # p.s.d., but there is nothing like this in spla (in Matlab there is a ichol)
    time0 = time.time()
    invA_approx = spla.spilu(A, drop_tol=1e-5,fill_factor=1e2)
    print('    Computing preconditioner: ' + str(time.time() - time0))
    
    time0 = time.time()
    (z, info) = spla.cg(A, b, M=spla.LinearOperator(A.shape, matvec=invA_approx.solve))
    if (info != 0) :
        print('WARNING: cg returned and info of ' + str(info))
    print('    Vergangene Zeit (Lösen LGS): ' + str(time.time() - time0))
    
    return z

################################################################################


def solve_qr(A, b) :
    """
    Solve linear least-squares problem by QR factorization
    
    A         One linear least-squares term per row, of size (Mh + Mg) x N, 
            with N the number of unknown elevation values and Mh the number of
            elevation terms and Mg the number of smoothing terms
    b         Right hand side, of size (Mh + Mg)
    """

   
    # Solve the least-squares problem by QR factorization. Don't head for the
    # normal euqations, as this squares the condition number. Moreover, solving
    # the arising system of linear equations is hard to solve when the dimension
    # increases (SuperLU and UMFPACK run out of memory for 1201x1201 tiles).
    #
    # The documentation of lsqr recommends a normalization of our matrix' 
    # columns to a Euclidean norm of 1. We have it here to play it safe. However
    # with proper smoothing lambda, scaling should not be crucial.
    A_normalizer = scipy.sparse.diags(np.squeeze(np.asarray(1. / np.sqrt(A.multiply(A).sum(0)))), 0, format=sparse_matrix_type)
    A = A * A_normalizer
    tol = 1e-8
    (z, lstop) = spla.lsqr(A, b, show=False, atol=tol, btol=tol, conlim=1e5)[0:2]
    z = A_normalizer * z
    
    return z


################################################################################


def solve_mr(A, b) :
    """
    Solve linear least-squares problem with lsmr
    
    A         One linear least-squares term per row, of size (Mh + Mg) x N, 
            with N the number of unknown elevation values and Mh the number of
            elevation terms and Mg the number of smoothing terms
    b         Right hand side, of size (Mh + Mg)
    """

    tol = 1e-8
    (z, lstop) = spla.lsmr(A, b, atol=tol, btol=tol, show=False)[0:2]
    
    return z


################################################################################

def unique(a):
    """
    This method deletes double-entries out of an array a. Comes from
    http://stackoverflow.com/questions/8560440/removing-duplicate-columns-and-rows-from-a-numpy-2d-array
    """
    
    # ignore elevation data, as we observed ASTER files for same lng,lat to 
    # have different elevation data. This is all no problem as we are just 
    # interested in removing x,y duplicates
    order = np.lexsort(a.T[0:2,:])
    a = a[order,:]
    diff = np.diff(a[:,0:2], axis=0)
    ui = np.ones(len(a), 'bool')
    # Don't test for equality but do an eps test instead. ASTER/SRTM x,y 
    # coordinates are computed and may suffer from numerical issues.
    ui[1:] = (np.absolute(diff) > eps).any(axis=1)
    return a[ui,:]


################################################################################  
    
    
def hoehenausgleich (X0,Y0,dx,dy,refSRTM,refASTER,refBEV):
    """
    This method builds the matrix A and vector b of the linear least squares
    problem w.r.t. the elevation data, \| A * z - b \|^2. Per reference 
    elevation we return a row in a sparse matrix A, along with a vector b.
    Hence, A is of shape (Nh x N), there Nh is the number of reference elevation
    data and N is the number of unknowns.
    
    For given elevation data z0, A * z0 - b gives the residual for each reference
    elevation data.
    """    

    #nullen=0
    #for i in range(len(refASTER)):
    #    if refASTER[i][3]==0:
    #        nullen+=1
    #print(nullen)
    
    
    
    dx1=np.round(np.trunc(dx)*3600+(dx-np.trunc(dx))*10000)   
    dy1=np.round(np.trunc(dy)*3600+(dy-np.trunc(dy))*10000)
    nrUnknowns_ASTER = int((dx1+1)*(dy1+1))
        
    
    nrSRTM  = len(refSRTM)
    nrASTER = len(refASTER)
    nrBEV   = len(refBEV)
    
    
    
    if (nrASTER != nrUnknowns_ASTER):
        print('ERROR: Unexpected number of ASTER points. We expected ' + str(nrUnknowns_ASTER) + ' points, but got ' + str(nrASTER))
        return
    
    rows = np.arange(nrSRTM + nrASTER + nrBEV)
    
    #print(shape(rows))
    # We must obtain the index for SRTM and ASTER points from multi_indexing
    # as unique(.) messed the initially nice ordering of this data.    
    cols_SRTM  = multi_indexing(X0,Y0,dx,dy,refSRTM)
    cols_ASTER = multi_indexing(X0,Y0,dx,dy,refASTER)
    
    cols_BEV   = np.arange(nrUnknowns_ASTER, nrUnknowns_ASTER + nrBEV)
    cols = np.concatenate((cols_SRTM, cols_ASTER, cols_BEV))
    
    ASTERweight = refASTER[:,3]
    
    data = np.concatenate((np.ones(nrSRTM), ASTERweight, np.tile(BEVweight, nrBEV)))
    
    if refBEV==[]:
        b = np.concatenate((refSRTM[:,2], ASTERweight*refASTER[:,2]))
    else:
        b = np.concatenate((refSRTM[:,2], ASTERweight*refASTER[:,2], BEVweight*refBEV[:,2]))
    
    water=refASTER[:,4]
    # build sparse matrix
    mask=np.concatenate((np.ones(nrSRTM, dtype=bool),np.logical_and((ASTERweight!=0),(water!=waterLabel)),np.tile(BEVweight != 0, nrBEV)))
    #mask=np.concatenate((np.ones(nrSRTM, dtype=bool),(ASTERweight!=0),np.ones(nrBEV, dtype=bool)))
    
    
    b=b[mask]
    data=data[mask]
    cols=cols[mask]
   
    rows=np.arange(shape(data)[0])    
    
    # We must give shape explicitly in case ASTERweight of the last element is 0.
    A = coo_matrix((data,(rows,cols)),shape=[len(b),nrASTER+nrBEV])
    A = coo_matrix.tocsc(A)
    #print(shape(A))
    
    return(A,b)
    

################################################################################



def glaettung (X0,Y0,dx,dy,refASTER,refBEV):
    """
    This method builds the matrix A and vector b of the linear least squares
    problem w.r.t. smoothing, \| A * z - b \|^2. Each row corresponds to smothing
    in a vertex of the Delaunay triangluation (a vertex corresponds to an unknown
    elevation value). A is of shape (Ng x N), where Ng is the number of vertices.
    
    For given elevation data z0, A * z0 - b gives the smoothing residual in each
    vertex.
    """    
    
    nrASTER_old = refASTER.shape[0]
    
    # re-order ASTER points such that the index of an ASTER point corresponds to
    # its variable index. unique(.) messes with the order of our SRTM/ASTER 
    # points
    cols_ASTER = multi_indexing(X0,Y0,dx,dy,refASTER)
    refASTER = refASTER[cols_ASTER,:]
    
    if (nrASTER_old != refASTER.shape[0]):
        print 'ERROR: glaettung reordering of ASTER points had a collision.'
        return
    
    #
    # compute delauny triangulation
    #
    
    datalist=[]
    datalist.append(refASTER[:,0:3])
    if not refBEV==[]:    
        datalist.append(refBEV)  
    dataN=np.vstack(datalist)
    n=shape(dataN)[0]
    D2dataN=dataN[:,0:2]
    D2dataN[:,0]*=isofaktor 
    # Does not change triangulation, but gives numerical stability
    D2dataN[:,0] -= X0;
    D2dataN[:,1] -= Y0;
    time0 = time.time();
    triang=scipy.spatial.Delaunay(D2dataN)
    D2dataN[:,0] += X0;
    D2dataN[:,1] += Y0;
    if (triang.coplanar.size != 0) :
        # Read http://www.qhull.org/html/qh-optq.htm#Qc
        print "glaettung: We have coplanar points."
        return
 
    print('    Delaunay Triangulation took ' + str(time.time() - time0) + ' seconds.')
    
    #
    # plot delaunay triangulation
    #
    
    if False : 
        plt.triplot(D2dataN[:,0], D2dataN[:,1], triang.simplices.copy())
        plt.plot(D2dataN[0:len(refASTER),0], D2dataN[0:len(refASTER),1], 'go')
        plt.plot(D2dataN[len(refASTER):len(D2dataN),0], D2dataN[len(refASTER):len(D2dataN),1], 'ro')
        plt.axis('equal')
        plt.show()
 
    #
    # initialize connectivity
    #            

    time0 = time.time();
    areas = findarea_ng(triang.simplices, D2dataN)
    print('    Computing areas of triangles adjacent to vertices took ' + str(time.time() - time0) + ' seconds.')

    time0 = time.time();
    edges_and_weights = computeGlaettungWeights(triang.simplices, D2dataN)
    print('    Computing edges and weights took ' + str(time.time() - time0) + ' seconds.')
        
    #
    # finally, build matrix and vector
    # 
    
    time0=time.time();
    bges=np.zeros(n)
    
    rows=[]
    cols=[]
    dat=[]
        
    rows.append(edges_and_weights[:,0])   
    rows.append(edges_and_weights[:,0])
    rows.append(edges_and_weights[:,1])
    rows.append(edges_and_weights[:,1])
     
    cols.append(edges_and_weights[:,0])   
    cols.append(edges_and_weights[:,1])
    cols.append(edges_and_weights[:,0])
    cols.append(edges_and_weights[:,1])
    
    dat.append( edges_and_weights[:,2])
    dat.append(-edges_and_weights[:,2])
    dat.append(-edges_and_weights[:,2])
    dat.append( edges_and_weights[:,2])
    
    rows = np.hstack(rows)
    cols = np.hstack(cols)
    dat  = np.hstack(dat)
    
    #print 'max(dat) = ' + str(np.amax(dat))
    #print 'min(dat) = ' + str(np.amin(dat))
            
    #print 'max(area) = ' + str(np.amax(areas))
    #print 'min(area) = ' + str(np.amin(areas))
    
    #print 'bring me in again'
    A = coo_matrix((dat,(rows,cols)), shape=(n,n))
    #A = coo_matrix((np.ones(len(dat)) * 1e8,(rows,cols)), shape=(n,n))
    A = coo_matrix.tocsc(A)
    
    # After this step, A is not symmetric any more
    D = scipy.sparse.diags(1. / areas, 0, format=sparse_matrix_type)
    #print 'bring me in again'
    A = A * D
    
    #Wir fordern für die Glättungsgewichte: Bei Asterweight=0 soll das Glättungs-
    #gewicht maximal sein. Bei ASTERweight=1 soll es 1 sein. Dies führt zu unten-
    #stehende Formeln.
        
    PARAM_maxglw=4
    
    #IndicesBEV weist jedem BEV-Punkt den nächstliegenden Index zu. Über diesen holen
    #wir uns num-Werte für BEV aus refASTER.    
    if not refBEV==[]: 
        indicesBEV=multi_indexing(X0,Y0,dx,dy,refBEV)
        BEVWeights = (1-PARAM_maxglw)*refASTER[indicesBEV,3]+PARAM_maxglw
    else:
		BEVWeights = [];
    ASTERweights = (1-PARAM_maxglw)*refASTER[:,3]+PARAM_maxglw;
    glw=np.concatenate((ASTERweights,BEVWeights))
    
    #Nun muss noch die Wassermaske berücksichtigt werden. 
    glw[refASTER[:,4]==waterLabel] = 2*PARAM_maxglw
    
    E = scipy.sparse.diags(glw, 0, format=sparse_matrix_type)
    #print 'bring me in again'
    A = A * E
    
    # Transpose, as we expect a term per row, and our terms are vertex-wise
    A = A.T
    
    print('    Assembling final matrix/vector took ' + str(time.time() - time0) + ' seconds.')
    
    return(A,bges)    



###############################################################################

def findarea_ng(simplices, points):
    """
    This method sums up the area of all triangles adjacent to a vertex. It avoids
    any for loops and thus we expect it to be faster than findarea. This awaits
    some benchmarking, however.
    
    On a 1201x1201 "Vorarlberg" tile, this method takes currently 1.4 seconds.    
    That's 110 times faster than legacy find_area. 
    """
    
    #
    # compute area per triangle
    #
    
    # from triang.simplices, compute two edge vectors for each triangle. 
    vec10 = points[simplices[:,1],:] - points[simplices[:,0],:]
    vec20 = points[simplices[:,2],:] - points[simplices[:,0],:]
    # np.cross returns just the z coordinate for 2D vectors! so area is now the
    # are of each triangle.
    area = np.abs(np.cross(vec10,vec20))*0.5
    
    #
    # compute sum of area of all triangles adjacent to a vertex 
    # 
    
    # we get (area0, area0, area0, area1, area1, area1, area2, area2, area2, ...) 
    areas = np.repeat(area, 3)
    # we get the three vertex indices of the first triangle, then the second, ...
    vindices = np.reshape(simplices, -1)
    n = points.shape[0]
    # this orders by vertex index and sums up duplicate entries for vertex 
    # indices. maybe there is a better way of doing this?
    vertex_area_sum = coo_matrix((areas,(vindices,vindices)), shape=(n,n)).diagonal()
    
    # compare 
    #    vertex_area_sum[i] 
    # against 
    #    np.sum(areas[vindices == i])
    
    return vertex_area_sum


###############################################################################


def computeGlaettungWeightsCot(points,ai,bi,ci) : 
    """
    Compute the cotanges of the angle at vertex c of triangle (a,b,c). Helper
    function for computeGlaettungWeights.
    """
    ac = points[ai.astype(int),:] - points[ci.astype(int),:];
    bc = points[bi.astype(int),:] - points[ci.astype(int),:];
    # compute norm of rows in ac and bc
    ac_norm = np.sqrt(np.sum(np.multiply(ac,ac), axis=1))
    bc_norm = np.sqrt(np.sum(np.multiply(bc,bc), axis=1))
    cosalpha = np.sum(np.multiply(ac,bc), axis=1) / np.multiply(ac_norm, bc_norm)
    return cosalpha / np.sqrt(1. - np.square(cosalpha))
    

##############################################################################


def computeGlaettungWeights(simplices, points) :
    """
    We compute the glaettung weights for each edge. We do so without any for-loop.
    we return an array of size (Ne x 4), where Ne is the Delaunay 
    triangulation's number of edges . The array stores a row per edge. The 
    first two columns give the vertex indices of the edge's end points. The index
    in the first column is always smaller than the one in the second column. The
    edges are not oriented, so there is only one entry per edge. The remaining
    column stores the edge weight.
    
        (v_i, v_j, w_ij)
    
    For 1 Mio udist points on [0,1]^2, this method takes 8 seconds
    For 5 Mio udist points on [0,1]^2, this method takes 46 seconds
    For udist points on [0,1]^2, this method takes 80-85% of the time of the
    Delaunay triangulation.
    
    simplices - Simplices attribute of the Delauny triangulation, a 
        nrTriangles x 3 array
    points - 2D point array of shape nrPoints x 2
    areas - Sum of areas of triangles at vertex i, an array of size nrTriangles.
    """
    
    nrTriangles = simplices.shape[0]
    # 'edges' will hold all interior edges twice, all boundary edges once. Each 
    # face contributes a row, all rows store their opposite vertex in the third 
    # column.
    edges = np.zeros((3*nrTriangles, 4))
    edges[0:nrTriangles,0] = np.minimum(simplices[:,0], simplices[:,1])
    edges[0:nrTriangles,1] = np.maximum(simplices[:,0], simplices[:,1])
    edges[0:nrTriangles,2] = simplices[:,2]
    
    edges[nrTriangles:2*nrTriangles,0] = np.minimum(simplices[:,0], simplices[:,2])
    edges[nrTriangles:2*nrTriangles,1] = np.maximum(simplices[:,0], simplices[:,2])
    edges[nrTriangles:2*nrTriangles,2] = simplices[:,1]
    
    edges[2*nrTriangles:3*nrTriangles,0] = np.minimum(simplices[:,1], simplices[:,2])
    edges[2*nrTriangles:3*nrTriangles,1] = np.maximum(simplices[:,1], simplices[:,2])
    edges[2*nrTriangles:3*nrTriangles,2] = simplices[:,0]
    
    # now, we do a lexicographic sort w.r.t. the first two colums. This will
    # order duplicate entries (= interior edges) next to each other. Boundary
    # edges will remain alone. See the source code of unique, from which we 
    # have this idea
    
    order = np.lexsort(edges.T[0:2,:])
    edges = edges[order,:]
    
    #
    # Identify boundary/interior edges 
    #
    
    # Diff has one row less than edges. Assume an additional (1,1) entry in 
    # the beginning. Duplicate rows in edges (= interior edges) will have zero 
    # entries, whereas boundary edges and first appearances of interior edges 
    # will have non-zero entries.
    diff = np.diff(edges[:,0:2], axis=0)

    isDupEdge = np.zeros(edges.shape[0], 'bool')
    isDupEdge[1:] = (diff == 0).all(axis=1)
    
    # boundary edges will have a -1 in the third column in the end
    edges[~isDupEdge,3] = -1;
    
    # This will write the duplicate edges' opposite vertex to the primary edge
    # as third column
    edges[isDupEdge[1:],3] = edges[isDupEdge,2]
    # boundaryEdges are those with fourth column == -1
    col3isminus1 = (edges[:,3] == -1)
    # This is a Nbx3 array, where Nb is the number of boundary edges
    boundaryEdges = edges[col3isminus1, 0:3]
    # This is a Nix3 array, where Ni is the number of interiors edges
    interiorEdges = edges[~col3isminus1 & ~isDupEdge, 0:4]

    #
    # Interior edges
    #
    
    # The edge is (a,b) and c and d are the opposite vertices. Alpha is the
    # angle at c, beta the angle at d.
    
    cotalpha = computeGlaettungWeightsCot(points, interiorEdges[:,0], interiorEdges[:,1], interiorEdges[:,2])
    cotbeta  = computeGlaettungWeightsCot(points, interiorEdges[:,0], interiorEdges[:,1], interiorEdges[:,3])
    
    interiorEdges[:,2] = (cotalpha + cotbeta) / 2;
    
    #
    # boundary edges
    #
    
    cotalpha = computeGlaettungWeightsCot(points, boundaryEdges[:,0], boundaryEdges[:,1], boundaryEdges[:,2])
    
    boundaryEdges[:,2] = cotalpha / 2;
    
    # vstack result and good bye
    
    edges_and_weights = np.vstack((interiorEdges[:,0:3], boundaryEdges))
    
    return edges_and_weights
    
    
##############################################################################    
    
    
def haircut(X0,Y0,dx,dy,data):
    """
    This method takes the input data (data, i.e. an n x 3-array) and checks, whether
    X and Y coordinates of the data are located within a certain area (specified by
    X0,Y0, coordinates of the lower left corner, and dx,dy,i.e. length and width of
    the area). All data outside of this area are thrown away.
    """
    X0=np.trunc(X0)+(X0-np.trunc(X0))*10000/3600
    Y0=np.trunc(Y0)+(Y0-np.trunc(Y0))*10000/3600
    dx=np.trunc(dx)+(dx-np.trunc(dx))*10000/3600
    dy=np.trunc(dy)+(dy-np.trunc(dy))*10000/3600
    
    index = (data[:,0]>X0-eps) & (data[:,0]<(X0+dx+eps)) & (data[:,1]>Y0-eps) & (data[:,1]<(Y0+dy+eps))
    data = data[index,:]
    
    return data
    

################################################################################


def multi_indexing (X0,Y0,dx,dy,refPoints):
    """    
    This method avoids multiple calls to indexing from hoehenausgleich. Instead,
    we convert an entire bunch of coordinates to indices. The performance gain
    is huge. By using this, hoehenausgleich on a 1201x1201 raster takes 0.11 
    instead of 84 seconds. That is a speed up of more than 750.
    """     
    X0=np.round(np.trunc(X0)*3600+(X0-np.trunc(X0))*10000)
    Y0=np.round(np.trunc(Y0)*3600+(Y0-np.trunc(Y0))*10000)    
    dx=np.round(np.trunc(dx)*3600+(dx-np.trunc(dx))*10000)   
    dy=np.round(np.trunc(dy)*3600+(dy-np.trunc(dy))*10000)    
    
    i = np.round(3600.*refPoints[:,0]-X0)
    j = np.round(Y0+dy-3600.*refPoints[:,1])
    n = (i+(dx+1)*j).astype(int)
    return(n)
    

################################################################################


def findelev (X,Y,filename):
    
    Z=fileloader(filename)
    X0=np.trunc(Z[0][0])+(Z[0][0]-np.trunc(Z[0][0]))*3600/10000
    Y0=np.trunc(Z[len(Z)-1][1])+(Z[len(Z)-1][1]-np.trunc(Z[len(Z)-1][1]))*3600/10000
    dx=Z[len(Z)-1][0]-Z[0][0]
    dxu=np.trunc(dx)+(dx-np.trunc(dx))*3600/10000
    dy=Z[0][1]-Z[len(Z)-1][1]
    dyu=np.trunc(dy)+(dy-np.trunc(dy))*3600/10000
    n=indexing(X0,Y0,dxu,dyu,X,Y)
    print (n)
    if ('srtm' or 'SRTM') in filename:
        n=(n%3601)/3 + (((n-(n%3601))/3601)/3)*1201
    print Z[n][2]
    return (Z[n][2])
    
    

################################################################################


def filefinder_BEV(X0,Y0,dx,dy):
    """
    This method gets a certain area as input (specified by X0,Y0, coordinates
    of the lower left corner, and dx,dy,i.e. length and width of the area). The
    method then finds all BEV-files, that at least partially cover this area and
    gives them back in a list.
    """
    X0=np.trunc(X0)+(X0-np.trunc(X0))*10000/3600
    Y0=np.trunc(Y0)+(Y0-np.trunc(Y0))*10000/3600
    dx=np.trunc(dx)+(dx-np.trunc(dx))*10000/3600
    dy=np.trunc(dy)+(dy-np.trunc(dy))*10000/3600
    
    minmaxmatrix=open("minmaxmatrix.txt")

    x=[]

    for line in minmaxmatrix.readlines():
        # Splits the read line by separating spaces into a list of string tokens 
        # and converts them to float
        y=line.split()       
        y[1:5] = [float(yy) for yy in y[1:5]]
        # add this rpw to global list
        x.append( y )

    minmaxmatrix.close()
    
    files=[]    
    
    for i in range(len(x)):
        if (x[i][1]>X0) and (x[i][1]<(X0+dx)) and (x[i][3]>Y0) and (x[i][3]<(Y0+dy)):
           files.append(x[i][0])
        elif (x[i][1]>X0)and(x[i][1]<(X0+dx))and(x[i][4]>Y0)and(x[i][4]<(Y0+dy)):
           files.append(x[i][0])           
        elif (x[i][2]>X0)and(x[i][2]<X0+dx)and(x[i][3]>Y0)and(x[i][3]<(Y0+dy)):
           files.append(x[i][0])           
        elif (x[i][2]>X0)and(x[i][2]<X0+dx)and(x[i][4]>Y0)and(x[i][4]<(Y0+dy)):
           files.append(x[i][0])
        elif (X0>x[i][1])and(X0<x[i][2])and(Y0>x[i][3])and(Y0<x[i][4]):
           files.append(x[i][0])
        elif ((X0+dx)>x[i][1])and((X0+dx)<x[i][2])and(Y0>x[i][3])and(Y0<x[i][4]):   
           files.append(x[i][0])
        elif (X0>x[i][1])and(X0<x[i][2])and((Y0+dy)>x[i][3])and((Y0+dy)<x[i][4]):  
           files.append(x[i][0])
        elif ((X0+dx)>x[i][1])and((X0+dx)<x[i][2])and((Y0+dy)>x[i][3])and((Y0+dy)<x[i][4]): 
           files.append(x[i][0])
           
    #print(files)
    
    return files    


################################################################################
    
    
def filefinder_ASTER_SRTM(X0,Y0,dx,dy):
    """
    This method gets a certain area as input (specified by X0,Y0, coordinates
    of the lower left corner, and dx,dy,i.e. length and width of the area). The
    method then finds all ASTER- and SRTM-files, that at least partially cover this
    area and gives them back in a list.
    """    
    filenames=[]
    
    X0=np.trunc(X0)+(X0-np.trunc(X0))*10000/3600
    Y0=np.trunc(Y0)+(Y0-np.trunc(Y0))*10000/3600
    dx=np.trunc(dx)+(dx-np.trunc(dx))*10000/3600
    dy=np.trunc(dy)+(dy-np.trunc(dy))*10000/3600
    
    
    
    XB=int(math.floor(X0))
    XE=int(math.floor(X0+dx))
    YB=int(math.floor(Y0))
    YE=int(math.floor(Y0+dy))
   
    for i in range(XB,XE+1):
        for j in range(YB,YE+1):    
            filename='N'+(str(j)).zfill(2)+'E'+(str(i)).zfill(3)
            filenames.append(filename)
         
    return filenames  


################################################################################
     
     
def fileloader(filename) :
    """
    There are three supported filetypes: 
    - hgt files (SRTM/ASTER and their resp.  weights from the filter)
    - GeoTIFFs (ASTER and the oe3d result files)
    - xyz file (BEV)
    We check the filename ending.
    """    
    filename_ending = filename[-4:].upper();
    if ( filename_ending == '.HGT' ) :
        return fileloader_HGT(filename, True)
    elif ( filename_ending == '.XYZ' ) :
        return fileloader_XYZ(filename)
    elif ( filename_ending == '.TIF' ) : 
        return fileloader_TIF(filename, True )
    else :
        print 'ERROR: Unknown filename ending in', filename
    
    return []  


################################################################################


def fileloader_HGT(filename, formatAsXYZ) :
    """    
    GDAL opens HGT files only from "NxxEyyy.hgt" named files. Our files are
    more generally named, have this as substring however. So we extract the
    required substring and create a symbolic link, that we load.
    """        
    match = re.search('N[0-9]{2}E[0-9]{3}', filename)
    if match is None :
        print 'ERROR: Failed to regexp ', filename
        return []
    
    # build name of symbolic link. We include the process ID in the path
    # as otherwise we get problems in parallel environments.
    linkdir = '/tmp/' + str(os.getpid());
    try:
        os.makedirs(linkdir)
    except:
        # That's a python nop 
        pass
    
    linkname = linkdir  + '/' + match.group(0) + '.hgt'
    
    # Check whether it is safe to create the symbolic link
    if os.path.islink(linkname) :
        os.remove(linkname)
    elif os.path.isfile(linkname) :
        print 'ERROR: There exists a real file with name ', linkname, '. So we cannot setup a symbolic link.'
        return []
    
    # create symbolic link
    os.symlink(filename, linkname)
    
    # open file with GDAL 
    gdal_dataset = gdal.Open(linkname, GA_ReadOnly)
    
    # remove symbolic link
    os.remove(linkname)
    
    if gdal_dataset is None:
        print 'ERROR: GDAL failed to open ', linkname, '/', filename
        return []
    
    return fileloader_process_GDAL_data(gdal_dataset, formatAsXYZ)


################################################################################


def fileloader_TIF(filename, formatAsXYZ) :
    """
    Load GeoTIFFs via GDAL. This is rather straight-forward, for post-processing
    the data into our storage scheme we use the same track as the HGT variant
    of fileloader.
    """
    
    # open file with GDAL 
    gdal_dataset = gdal.Open(filename, GA_ReadOnly)
        
    if gdal_dataset is None:
        print 'ERROR: GDAL failed to open ', filename
        return []
    
    return fileloader_process_GDAL_data(gdal_dataset, formatAsXYZ)


################################################################################


def fileloader_process_GDAL_data(gdal_dataset, formatAsXYZ) :
    """
    Convert GDAL elevation data into our storage scheme
    """
    
    # read data into a XSize*YSize long list of signed integers    
    gdal_band = gdal_dataset.GetRasterBand(1)
    gdal_data_raw = gdal_band.ReadRaster( 0, 0, gdal_band.XSize, gdal_band.YSize, gdal_band.XSize, gdal_band.YSize, GDT_Int16)
    gdal_data = struct.unpack('h' * (gdal_band.XSize * gdal_band.YSize), gdal_data_raw)
    
    # now transform data into our target format
    # This yields a Nx1 matrix
    # We need floats as we store longitude/latitude along with the elevation
    data = np.transpose(np.asarray([gdal_data], dtype=np.float))
    
    if ( formatAsXYZ ) :
        gdal_geotransform = gdal_dataset.GetGeoTransform()
        if gdal_geotransform is None:
            print 'ERROR: GDAL failed to obtain geotransform from  ', linkname, '/', filename
            return []    
        # x is longitude, y is latitude. (x0,y0) is the top-left corner
        deltax = gdal_geotransform[1]
        deltay = gdal_geotransform[5]
        # Aster stores the corner of the pixel, our elevation value is located
        # in the pixel's center. Hence, we shift by deltax/deltay. deltay
        # is negative.
        x0 = (gdal_geotransform[0] + 0.5*deltax)
        y0 = (gdal_geotransform[3] + 0.5*deltay)    
        
        # deltax and deltay have alternating sign
        if (abs(deltax + deltay) > eps) :
            print 'ERROR: Data is not on a square coordinate grid.'
            return []    
        
        single_x = np.arange(start=x0, stop=x0+deltax*(gdal_band.XSize-0.5), step=deltax)
        single_y = np.arange(start=y0, stop=y0+deltay*(gdal_band.YSize-0.5), step=deltay)
        
        
        # We round, otherwise unqiue might mess us up
        single_x = np.around(single_x, decimals=8)
        single_y = np.around(single_y, decimals=8)
        
        x, y = np.meshgrid(single_x, single_y)
        
        x = np.reshape(x, (gdal_band.XSize*gdal_band.YSize,1), 'C');
        y = np.reshape(y, (gdal_band.XSize*gdal_band.YSize,1), 'C');
           
        # put everything into a Nx3 matrix
        data = np.concatenate((x,y,data),1)
        
    return data

################################################################################


def fileloader_XYZ(filename) :
    """
    Loads data from XYZ files, that is files with three space sperated columns
    """
    fileXYZ=open(filename)

    data=[]

    # I assume it is faster to collect in a list of 3D arrays and then convert
    # the list to a matrix in the end.
    for line in fileXYZ.readlines() :
        # see http://stackoverflow.com/questions/11747125/python-numpy-convert-string-in-to-numpy-array
        data.append(np.array(line.split(" "), dtype=np.float))

    fileXYZ.close()
    
    # convert list to Nx3 matrix. Use data.shape to get the sizes, use
    # data[0:3,:] to get the first three lines, etc.
    data=np.vstack(data)
    
    return data


################################################################################


def fileexporter(X0,Y0,dx,dy,z,filename):
    """
    Writes the elevation data "z" on the work window (X0,Y0,dx,dy) to the file with
    name "filename". From the filename we extract the extension that defines
    the file format. The elevation data must be odererd according to our linear
    indexing rules, e.g. the first entry must the the top-left pixel and so on.
    
    The numbers defining the work window (X0,Y0,dx,dy) are of special format.
    The integer part defines the degrees. The fractional part shall be 
    interpreted as integer defining the arc seconds. So 46.0099 means "46 
    degrees, 99 arc seonds". This holds for all four numbers.
    
    (X0,Y0) is the lower left corner of the working window.
    
    We support GeoTIFF export only, as we have metadata with copyright 
    information there. Use gdal command line tools to convert to 
    HGT.
    """
    X0_float = np.trunc(X0) + (X0-np.trunc(X0))*10000/3600
    Y0_float = np.trunc(Y0) + (Y0-np.trunc(Y0))*10000/3600
    dy_float = np.trunc(dy) + (dy-np.trunc(dy))*10000/3600
    nrCols = int(np.round(np.trunc(dx)*3600 + (dx-np.trunc(dx))*10000))+1
    nrRows = int(np.round(np.trunc(dy)*3600 + (dy-np.trunc(dy))*10000))+1
        
    # check number of elevation data vs. size of working window
    if (nrRows * nrCols != len(z)):
        print 'ERROR: The length of the elevation data ', len(z), ' does not match the work window size of ', nrRows, 'x', nrCols, '.'
        return False
    
    # get filename ending. For a list of supported file formats and their 
    # namings in GDAL, see http://www.gdal.org/formats_list.html.
    fileformat = ""
    filename_ending = filename[-4:].upper();
    if ( filename_ending == '.TIF' ) :
        fileformat = "GTiff"
    else:
        print 'ERROR: Unknown filename ending in', filename
        return False
       
    # Get driver that will write our raster data to a file. If not specified
    # otherwise, all the following is from
    #     http://www.gdal.org/gdal_tutorial.html.
    gdal_driver = gdal.GetDriverByName(fileformat)
    if (gdal_driver is None):
        print 'ERROR: Unsupported GDAL driver ', fileformat
        return False
        
    # Create a GDAL dataset
    gdal_dataset = gdal_driver.Create(filename, nrCols, nrRows, 1, gdal.GDT_Int16)
    
    licenseTag = 'Rechenraum e.U., based on data (c) BEV, 2012, (c) Aster, a product of METI and NASA';
    if (waterpfad != ''):
        licenseTag += ', and (c) OpenStreetMap';
    
    # write meta data. Syntax extracted from 
    #     https://gist.github.com/schwehr/5016127.
    # For a list of supported GDAL / GeoTIFF Metadata see
    #     http://www.gdal.org/frmt_gtiff.html
    gdal_dataset.SetMetadata({
        'TIFFTAG_ARTIST': 'Rechenraum e.U. (http://www.rechenraum.com)',
        'TIFFTAG_COPYRIGHT' : licenseTag,
        'TIFFTAG_DOCUMENTNAME' : 'oe3d',
        'TIFFTAG_DATETIME' :  datetime.datetime.now().isoformat()
    })
    
    # Set affine mapping (tx, r00, r01, ty, r10, r11), where (tx,ty) is the 
    # translation and (r00, r01; r10, r11) is the rotational matrix.
    # To verify the mapping, work with gdal_info on the result file.
    # (tx,ty) is the upper left corner of the upper left pixel.
    # See http://www.gdal.org/classGDALDataset.html#af9593cc241e7d140f5f3c4798a43a668.
    gdal_dataset.SetGeoTransform( [ X0_float-1./7200, 1./3600, 0, Y0_float+dy_float+1./7200, 0, -1./3600 ] )
    
    # see http://www.gdal.org/ogr/osr_tutorial.html
    srs = osr.SpatialReference()
    srs.SetWellKnownGeogCS("WGS84")
    gdal_dataset.SetProjection(srs.ExportToWkt())

    # reshape is row major, see np.arange(1,13).reshape(3,4)
    gdal_dataset.GetRasterBand(1).WriteArray(z.reshape(nrRows, nrCols))

    # Once we're done, close the dataset properly. This writes the file.
    gdal_dataset = None


################################################################################

def fileexporter_obj (dx,dy,ASTER,z,filename):
    
    
    laenge=(ASTER[1][0]+ASTER[len(ASTER)-1][0]+ASTER[len(ASTER)-dx][0]+ASTER[dx][0])/4
    breite=(ASTER[1][1]+ASTER[len(ASTER)-1][1]+ASTER[len(ASTER)-dx][1]+ASTER[dx][1])/4
    zm=(z[0]+z[dx]+z[len(z)-dx]+z[len(z)-1])/4    
    #print laenge, breite, zm    
    for i in range(len(ASTER)):
        ASTER[i][0]=((ASTER[i][0])-laenge)*2*pi*6371000*isofaktor/360
        ASTER[i][1]=((ASTER[i][1])-breite)*2*pi*6371000/360
        z[i]=z[i]-zm    
    #ASTER-Werte in m umrechnen!
    dx=int(np.trunc(dx)*3600+(dx-np.trunc(dx))*10000)
    dy=int(np.trunc(dy)*3600+(dy-np.trunc(dy))*10000)
    file=open(filename,"w")
    file.write('mtllib '+filename+'.mtl\n')    
    for i in range(len(ASTER)):
        file.write("v "+str(ASTER[i][0])+" "+str(ASTER[i][1])+" "+str(z[i])+"\n")
    for m in range(dy+1):
        for l in range(dx+1):
            file.write("vt "+str(float(l)/float(dx))+" "+str((dy-float(m))/dy)+" "+"0\n")
    file.write('usemtl luftbild\n')
    for j in range(dy):
        for k in range(dx):
            f0=str(k+1+j*(dx+1))
            f1=str(k+2+j*(dx+1))
            f2=str(k+2+(j+1)*(dx+1))          
            f3=str(k+1+(j+1)*(dx+1))
            file.write("f "+f0+"/"+f0+" "+f1+"/"+f1+" "+f2+"/"+f2+" "+f3+"/"+f3+"\n")
    file.close()
    
    #file = open("/home/gerhard/Desktop/data/newfile.txt", "w")
    
    #file.write("hello world in the new file\n")

    #file.write("and another line\n")

    #file.close()

################################################################################

# test code is now in compute.py
